"""Module containing custom types."""

from typing import Literal, TypedDict, Union
from typing_extensions import Required, NotRequired

from numpy import floating
from numpy.typing import NDArray

SubjectTimeseries = dict[str, dict[str, NDArray[floating]]]
"""
   Type Definition for the Subject Timeseries Dictionary Structure.

   A dictionary mapping subject IDs to their run IDs and their associated timeseries (TRs x ROIs) as a NumPy array.
   The structure is as follows:

    ::

        subject_timeseries = {
                "101": {
                    "run-0": np.array(shape=[TR, ROIs]),
                    "run-1": np.array(shape=[TR, ROIs]),
                    "run-2": np.array(shape=[TR, ROIs]),
                },
                "102": {
                    "run-0": np.array(shape=[TR, ROIs]),
                    "run-1": np.array(shape=[TR, ROIs]),
                }
            }

    .. important:: The run IDs must be in the form "run-{0}" (e.g. "run-0" or "run-zero").
"""


class SchaeferParcelConfig(TypedDict):
    """
    Type Definition for the Schaefer Parcellation Configurations.

    A ``TypedDict`` representing the available subkeys (second level keys for "Schaefer") for initializing the Schaefer
    parcellation in the ``TimeseriesExtractor`` or ``CAP`` classes. The structure is as follows:

    ::

        {"n_rois": 400, "yeo_networks": 7, "resolution_mm": 1}

    Parameters
    ----------
    n_rois: :obj:`int`
        Number of ROIs (Default=400). Options are 100, 200, 300, 400, 500, 600, 700, 800, 900, or 1000.
    yeo_networks: :obj:`int`
        Number of Yeo networks (Default=7). Options are 7 or 17.
    resolution_mm: :obj:`int`
        Spatial resolution in millimeters (Default=1). Options are 1 or 2.

    See Also
    --------
    ParcelConfig
        Type definition representing the configuration options and structure for the Schaefer and AAL parcellations.

    Notes
    -----
    See `Nilearn's fetch Schaefer documentation\
    <https://nilearn.github.io/stable/modules/generated/nilearn.datasets.fetch_atlas_schaefer_2018.html>`_  for more information.
    """

    n_rois: NotRequired[int]
    yeo_networks: NotRequired[int]
    resolution_mm: NotRequired[int]


class AALParcelConfig(TypedDict):
    """
    Type Definition for the AAL Parcellation Configurations.

    A ``TypedDict`` representing the available subkeys (second level keys for "AAL") for initializing the AAL
    parcellation in the ``TimeseriesExtractor`` or ``CAP`` classes. The structure is as follows:

    ::

        {"version": "SPM12"}

    Parameters
    ----------
    version: :obj:`str`
        AAL parcellation version to use (Default="SPM12" if ``{"AAL": {}}`` is given). Options are "SPM5", "SPM8",
        "SPM12", or "3v2".

    See Also
    --------
    ParcelConfig
        Type definition representing the configuration options and structure for the Schaefer and AAL parcellations.

    Notes
    -----
    See `Nilearn's fetch AAL documentation\
    <https://nilearn.github.io/stable/modules/generated/nilearn.datasets.fetch_atlas_aal.html>`_ for more information.
    """

    version: NotRequired[str]


ParcelConfig = Union[dict[Literal["Schaefer"], SchaeferParcelConfig], dict[Literal["AAL"], AALParcelConfig]]
"""
   Type Definition for the Parcellation Configurations.

   A dictionary mapping the Schaefer or AAL parcellation to their associated configuration subkeys that are used
   by the ``TimeseriesExtractor`` and ``CAP`` class to create the processed ``ParcelApproach``. The structure is
   as follows:

    ::

        # Structure of Schaefer
        {"Schaefer": SchaeferParcelConfig}


        # Structure of AAL
        {"AAL": AALParcelConfig}

    See Also
    --------
    :class:`neurocaps.typing.SchaeferParcelConfig`
        Type definition representing configuration options for the Schaefer parcellation.
    :class:`neurocaps.typing.AALParcelConfig`
        Type definition representing configuration options for the AAL parcellation.
"""


# No doc string
class ParcelApproachBase(TypedDict):
    maps: NotRequired[str]
    nodes: NotRequired[list[str]]


class SchaeferParcelApproach(ParcelApproachBase):
    """
    Type Definition for the Schaefer Parcellation Approach.

    A ``TypedDict`` representing the subkeys (second level keys for "Schaefer") for the processed Schaefer parcellation
    produced by the ``TimeseriesExtractor`` or ``CAP`` classes. The structure is as follows:

    ::

        {
            "maps": "path/to/parcellation.nii.gz",
            "nodes": ["LH_Vis1", "LH_SomSot1", "RH_Vis1", "RH_Somsot1"],
            "regions": ["Vis", "SomSot"]
        }

    Parameters
    ----------
    maps: :obj:`str`
        Path to the Schaefer parcellation.
    nodes: :obj:`list[str]`
        List of nodes (ROIs) in the Schaefer parcellation. Ordered in ascending order of their label ID in the
        parcellation and must exclude "Background".
    regions: :obj:`list[str]`
        List of networks in the Schaefer parcellation. **Important**: For certain visualization methods, the ``in``
        operator is used to determine which nodes belong to which network. Therefore, network names must be contained
        within the corresponding node names (e.g., 'Vis' network should have nodes with 'Vis' in their names).

    See Also
    --------
    ParcelApproach
        Type definition representing the structure of the Schaefer, AAL, and Custom parcellation approaches.
    """

    regions: NotRequired[list[str]]


class AALParcelApproach(ParcelApproachBase):
    """
    Type Definition for the AAL Parcellation Approach.

    A ``TypedDict`` representing the subkeys (second level keys for "AAL") for the processed AAL parcellation produced
    by the ``TimeseriesExtractor`` or ``CAP`` classes. The structure is as follows:

    ::

        {
            "maps": "path/to/parcellation.nii.gz",
            "nodes": ["Precentral_L", "Precentral_R", "Frontal_Sup", "Frontal_Sup_R"],
            "regions": ["Precentral", "Frontal"]
        }

    Parameters
    ----------
    maps: :obj:`str`
        Path to the AAL parcellation.
    nodes: :obj:`list[str]`
        List of nodes (ROIs) in the AAL parcellation. Ordered in ascending order of their label ID in the parcellation
        and must exclude "Background".
    regions: :obj:`list[str]`
        List of networks in the AAL parcellation. **Important**: For certain visualization methods, the ``in``
        operator is used to determine which nodes belong to which network. Therefore, network names must be contained
        within the corresponding node names (e.g., 'Frontal' network should have nodes with 'Frontal' in their names).

    See Also
    --------
    ParcelApproach
        Type definition representing the structure of the Schaefer, AAL, and Custom parcellation approaches.
    """

    regions: NotRequired[list[str]]


class CustomRegionHemispheres(TypedDict):
    """
    Type Definition for Hemisphere Mapping in Custom Parcellation Regions.

    A ``TypedDict`` representing the mapping of the index position of the "nodes" to the left and right hemispheres.

    ::

        {"lh": [0, 1], "rh": [3, 4, 5]}

    Parameters
    ----------
    lh: :obj:`list[int] | range`
       List of integers or range representing the index positions of elements in the "nodes" list belonging to the
       left hemisphere of a specific region.
    rh: :obj:`list[int] | range`
       List of integers or range representing the index positions of elements in the "nodes" list belonging to the
       right hemisphere of a specific region.

    See Also
    --------
    CustomParcelApproach
       The type definition for the Custom parcellation approach.
    """

    lh: Required[Union[list[int], range]]
    rh: Required[Union[list[int], range]]


class CustomParcelApproach(ParcelApproachBase):
    """
    Type Definition for the Custom Parcellation Approach.

    A ``TypedDict`` representing the subkeys (second level keys for "Custom") for the user-defined Custom parcellation
    approach. The structure is as follows:

    ::

        {
            "maps": "path/to/parcellation.nii.gz",
            "nodes": ["LH_Vis1", "LH_Vis2", "LH_Hippocampus", "RH_Vis1", "RH_Vis2", "RH_Vis3", "RH_Hippocampus"],
            "regions": {
                "Visual": CustomRegionHemispheres
                "Hippocampus": CustomRegionHemispheres
            }
        }

    Parameters
    ----------
    maps: :obj:`str`
        Path to the Custom parcellation.
    nodes: :obj:`list[str]`
        List of nodes (ROIs) in the Custom parcellation. Ordered in ascending order of their label ID in the
        parcellation and must exclude "Background".
    regions: :obj:`dict[str, CustomRegionHemispheres]`
        Dictionary mapping the regions to their left and right hemispheres.

    See Also
    --------
    CustomRegionHemispheres
        Type definition of the Custom hemisphere dictionary for the "regions" subkeys.
    ParcelApproach
        Type definition representing the structure of the Schaefer, AAL, and Custom parcellation approaches.
    """

    regions: NotRequired[dict[str, CustomRegionHemispheres]]


ParcelApproach = Union[
    dict[Literal["Schaefer"], SchaeferParcelApproach],
    dict[Literal["AAL"], AALParcelApproach],
    dict[Literal["Custom"], CustomParcelApproach],
]
"""
  Type Definition for the Parcellation Approaches.

   A dictionary mapping the Schaefer, AAL, and Custom parcellation approaches to their associated subkeys:

    ::

        # Structure of Schaefer
        {"Schaefer": SchaeferParcelApproach}

        # Structure of AAL
        {"AAL": AALParcelApproach}

        # Structure of Custom
        {"Custom": CustomParcelApproach}

    See Also
    --------
    :class:`neurocaps.typing.SchaeferParcelApproach`
        Type definition representing the structure of the Schaefer parcellation approach.
    :class:`neurocaps.typing.AALParcelApproach`
        Type definition representing the structure of the AAL parcellation approach.
    :class:`neurocaps.typing.CustomParcelApproach`
        Type definition representing the structure of the Custom parcellation approach.
"""

__all__ = [
    "SubjectTimeseries",
    "SchaeferParcelConfig",
    "AALParcelConfig",
    "ParcelConfig",
    "SchaeferParcelApproach",
    "AALParcelApproach",
    "CustomParcelApproach",
    "CustomRegionHemispheres",
    "ParcelApproach",
]



"""A class which is responsible for accessing attributes in ``CAP``."""

import copy, sys
from typing import Union

import numpy as np
from numpy.typing import NDArray
from sklearn.cluster import KMeans

from ..check_parcel_approach import _check_parcel_approach
from ...typing import ParcelConfig, ParcelApproach


class _CAPGetter:
    def __init__(self):
        pass

    ### Attributes exist when CAP initialized
    @property
    def parcel_approach(self) -> Union[ParcelApproach, None]:
        return self._parcel_approach

    @parcel_approach.setter
    def parcel_approach(self, parcel_dict: Union[ParcelConfig, ParcelApproach, str]) -> None:
        self._parcel_approach = _check_parcel_approach(parcel_approach=parcel_dict, call="setter")

    @property
    def groups(self) -> Union[dict[str, list[str]], None]:
        return self._groups

    ### Attributes exist when CAP.get_caps() used
    @property
    def subject_table(self) -> Union[dict[str, str], None]:
        return getattr(self, "_subject_table", None)

    @subject_table.setter
    def subject_table(self, subject_dict: dict[str, str]) -> None:
        if isinstance(subject_dict, dict):
            self._subject_table = copy.deepcopy(subject_dict)
        else:
            raise TypeError(
                "`self.subject_table` must be a dictionary where the keys are the subject IDs and the "
                "values are the group names."
            )

    @property
    def n_clusters(self) -> Union[int, list[int], None]:
        return getattr(self, "_n_clusters", None)

    @property
    def cluster_selection_method(self) -> Union[str, None]:
        return getattr(self, "_cluster_selection_method", None)

    @property
    def n_cores(self) -> Union[int, None]:
        return getattr(self, "_n_cores", None)

    @property
    def runs(self) -> Union[list[Union[int, str]], None]:
        return getattr(self, "_runs", None)

    @property
    def standardize(self) -> Union[bool, None]:
        return getattr(self, "_standardize", None)

    @property
    def means(self) -> Union[dict[str, NDArray[np.floating]], None]:
        return getattr(self, "_mean_vec", None)

    @property
    def stdev(self) -> Union[dict[str, NDArray[np.floating]], None]:
        return getattr(self, "_stdev_vec", None)

    @property
    def concatenated_timeseries(self) -> Union[dict[str, NDArray[np.floating]], None]:
        return getattr(self, "_concatenated_timeseries", None)

    @concatenated_timeseries.deleter
    def concatenated_timeseries(self) -> None:
        del self._concatenated_timeseries

    @property
    def kmeans(self) -> Union[dict[str, KMeans], None]:
        return getattr(self, "_kmeans", None)

    @property
    def caps(self) -> Union[dict[str, dict[str, NDArray[np.floating]]], None]:
        return getattr(self, "_caps", None)

    @property
    def cluster_scores(self) -> Union[dict[str, Union[str, dict[str, float]]], None]:
        return getattr(self, "_cluster_scores", None)

    @property
    def optimal_n_clusters(self) -> Union[dict[str, int], None]:
        return getattr(self, "_optimal_n_clusters", None)

    @property
    def variance_explained(self) -> Union[dict[str, float], None]:
        return getattr(self, "_variance_explained", None)

    # Generated in `caps2plot`
    @property
    def region_means(self) -> Union[dict[str, dict[str, Union[list[str], NDArray[np.floating]]]], None]:
        return getattr(self, "_region_means", None)

    @property
    def outer_products(self) -> Union[dict[str, dict[str, NDArray[np.floating]]], None]:
        return getattr(self, "_outer_products", None)

    # Generated in `caps2radar`
    @property
    def cosine_similarity(self) -> Union[dict[str, dict[str, Union[list[str], NDArray[np.floating]]]], None]:
        return getattr(self, "_cosine_similarity", None)

    def _concatenated_timeseries_size(self) -> str:
        if not self.concatenated_timeseries:
            return "0 bytes"

        total_bytes = sum(arr.nbytes for arr in self.concatenated_timeseries.values())
        total_bytes += sys.getsizeof(self.concatenated_timeseries)

        return f"{total_bytes} bytes"

    def __str__(self) -> str:
        parcellation_name = list(self.parcel_approach.keys())[0] if self.parcel_approach else None
        # Get group names
        groups_names = ", ".join(f"{k}" for k in self.groups.keys()) if self.groups else None
        # Get total CAPs per group
        group_caps = {k: len(v) for k, v in self.caps.items()} if self.caps else None

        object_properties = (
            f"Parcellation Approach                           : {parcellation_name}\n"
            f"Groups                                          : {groups_names}\n"
            f"Number of Clusters                              : {self.n_clusters}\n"
            f"Cluster Selection Method                        : {self.cluster_selection_method}\n"
            f"Optimal Number of Clusters                      : {self.optimal_n_clusters}\n"
            f"CPU Cores Used for Clustering (Multiprocessing) : {self.n_cores}\n"
            f"User-Specified Runs IDs Used for Clustering     : {self.runs}\n"
            f"Concatenated Timeseries Bytes                   : {self._concatenated_timeseries_size()}\n"
            f"Standardized Concatenated Timeseries            : {self.standardize}\n"
            f"Co-Activation Patterns (CAPs)                   : {group_caps}\n"
            f"Variance Explained by Clustering                : {self.variance_explained}"
        )

        sep = "=" * len(object_properties.rsplit(": ")[0])

        return "Metadata:\n" + sep + f"\n{object_properties}"



"""A class which is responsible for accessing attributes in ``TimeSeriesExtractor``."""

import copy, sys
from typing import Union

import numpy as np

from ..check_parcel_approach import _check_parcel_approach
from ..pickle_utils import _convert_pickle_to_dict
from ...typing import ParcelConfig, ParcelApproach, SubjectTimeseries


class _TimeseriesExtractorGetter:
    def __init__(self):
        pass

    #### Exists upon initialization of TimeseriesExtractor
    @property
    def space(self) -> str:
        return self._space

    @space.setter
    def space(self, new_space: str) -> None:
        if not isinstance(new_space, str):
            raise TypeError("`space` must be a string.")
        self._space = new_space

    @property
    def parcel_approach(self) -> ParcelApproach:
        return self._parcel_approach

    @parcel_approach.setter
    def parcel_approach(self, parcel_dict: Union[ParcelConfig, ParcelApproach, str]) -> None:
        self._parcel_approach = _check_parcel_approach(parcel_approach=parcel_dict, call="setter")

    @property
    def signal_clean_info(self) -> Union[dict[str, Union[bool, int, float, str]], None]:
        return self._signal_clean_info

    ### Does not exists upon initialization of Timeseries Extractor

    # Exist when TimeSeriesExtractor.get_bold() used
    @property
    def task_info(self) -> Union[dict[str, Union[str, int]], None]:
        return getattr(self, "_task_info", None)

    # Gets initialized and populated in TimeSeriesExtractor.get_bold(),
    @property
    def subject_ids(self) -> Union[list[str], None]:
        return getattr(self, "_subject_ids", None)

    @property
    def n_cores(self) -> Union[int, None]:
        return getattr(self, "_n_cores", None)

    # Gets initialized in TimeSeriesExtractor.get_bold(), gets populated when
    # TimeseriesExtractor._timeseries_aggregator gets called in TimeseriesExtractor._extract_timeseries
    @property
    def subject_timeseries(self) -> Union[SubjectTimeseries, None]:
        return getattr(self, "_subject_timeseries", None)

    @subject_timeseries.setter
    def subject_timeseries(self, subject_dict: Union[SubjectTimeseries, str]) -> None:
        need_deepcopy = True

        if isinstance(subject_dict, str) and subject_dict.endswith(".pkl"):
            subject_dict = _convert_pickle_to_dict(subject_dict)
            need_deepcopy = False

        self._validate_timeseries(subject_dict)

        if need_deepcopy:
            self._subject_timeseries = copy.deepcopy(subject_dict)
        else:
            self._subject_timeseries = subject_dict

    @subject_timeseries.deleter
    def subject_timeseries(self) -> None:
        del self._subject_timeseries

    @property
    def qc(self) -> Union[dict, None]:
        return getattr(self, "_qc", None)

    @staticmethod
    def _validate_timeseries(subject_dict: SubjectTimeseries) -> None:
        error_msg = (
            "A valid pickle file/subject timeseries should contain a nested dictionary where the "
            "first level is the subject id, second level is the run number in the form of 'run-#', and "
            "the final level is the timeseries as a numpy array. "
        )

        error_dict = {
            "Sub": error_msg + "The error occurred at [SUBJECT: {0}]. ",
            "Run": error_msg + "The error occurred at [SUBJECT: {0} | RUN: {1}]. ",
        }

        if not isinstance(subject_dict, dict):
            raise TypeError(error_msg)

        for sub in subject_dict:
            if not isinstance(subject_dict[sub], dict):
                raise TypeError(
                    error_dict["Sub"].format(sub) + "The subject must be a dictionary with second level "
                    "'run-#' keys."
                )

            runs = list(subject_dict[sub])

            if not all("run" in x for x in runs):
                raise TypeError(error_dict["Sub"].format(sub) + "Not all second level keys follow the form of 'run-#'.")

            for run in runs:
                if not isinstance(subject_dict[sub][run], np.ndarray):
                    raise TypeError(error_dict["Run"].format(sub, run) + "All 'run-#' keys must contain a numpy array.")

    def _subject_timeseries_size(self) -> str:
        if not self.subject_timeseries:
            return "0 bytes"

        total_bytes = sum(arr.nbytes for subject in self.subject_timeseries.values() for arr in subject.values())
        # Adding size of dictionary
        total_bytes += sys.getsizeof(self.subject_timeseries)

        return f"{total_bytes} bytes"

    def __str__(self) -> str:
        n_subjects = len(self.subject_ids) if self.subject_ids else None

        object_properties = (
            f"Preprocessed Bold Template Space                           : {self.space}\n"
            f"Parcellation Approach                                      : {list(self.parcel_approach.keys())[0]}\n"
            f"Signal Clean Information                                   : {self.signal_clean_info}\n"
            f"Task Information                                           : {self.task_info}\n"
            f"Number of Subjects                                         : {n_subjects}\n"
            f"CPU Cores Used for Timeseries Extraction (Multiprocessing) : {self.n_cores}\n"
            f"Subject Timeseries Byte Size                               : {self._subject_timeseries_size()}"
        )

        sep = "=" * len(object_properties.rsplit(": ")[0])

        return "Metadata:\n" + sep + f"\n{object_properties}"



"""Internal Function for checking confound names"""

from ..logger import _logger

LG = _logger(__name__)


def _check_confound_names(high_pass, user_confounds, n_acompcor_separate):
    """
    Pipeline for checking confound names. The default confound names ("basic") depend on whether high-pass filtering
    is specified.
    """
    if user_confounds == "basic":
        if high_pass:
            # Do not use cosine or acompcor regressor if high-pass filtering is not None.
            # Acompcor regressors are estimated on high pass filtered version of data form fmriprep
            confound_names = [
                "trans_x",
                "trans_x_derivative1",
                "trans_y",
                "trans_y_derivative1",
                "trans_z",
                "trans_z_derivative1",
                "rot_x",
                "rot_x_derivative1",
                "rot_y",
                "rot_y_derivative1",
                "rot_z",
                "rot_z_derivative1",
            ]
        else:
            confound_names = [
                "cosine*",
                "trans_x",
                "trans_x_derivative1",
                "trans_y",
                "trans_y_derivative1",
                "trans_z",
                "trans_z_derivative1",
                "rot_x",
                "rot_x_derivative1",
                "rot_y",
                "rot_y_derivative1",
                "rot_z",
                "rot_z_derivative1",
                "a_comp_cor_00",
                "a_comp_cor_01",
                "a_comp_cor_02",
                "a_comp_cor_03",
                "a_comp_cor_04",
                "a_comp_cor_05",
            ]
    else:
        assert isinstance(user_confounds, list) and user_confounds, "`confound_names` must be a non-empty list."
        confound_names = user_confounds

    if n_acompcor_separate:
        confound_names = _remove_a_comp_cor(confound_names, user_confounds, n_acompcor_separate)

    _check_regressors(confound_names, n_acompcor_separate)

    LG.info(f"Confound regressors to be used if available: {', '.join(confound_names)}.")

    return confound_names


def _remove_a_comp_cor(confound_names, user_confounds, n):
    """
    Removes all "a_comp_cor" regressors in ``confound_names`` if separate components for the white matter and
    cerebrospinal fluid masks are requested.
    """
    check_confounds = [confound for confound in confound_names if "a_comp_cor" not in confound]
    if len(confound_names) > len(check_confounds):
        removed_confounds = [element for element in confound_names if element not in check_confounds]
        if user_confounds:
            LG.warning(
                "Since `n_acompcor_separate` has been specified, acompcor components in "
                f"`confound_names` will be disregarded and replaced with the first {n} "
                "components of the white matter and cerebrospinal fluid masks for each participant. "
                f"The following components will not be used: {', '.join(removed_confounds)}."
            )

    return check_confounds


def _check_regressors(confound_names, n):
    """
    Performs a basic check to see if at least one "cosine" regressor is specified if "a_comp_cor" and "t_comp_cor"
    are detected in ``confound_names``.
    """
    cosine = any(i.startswith("cosine") for i in confound_names)
    acompcor = any(i.startswith("a_comp_cor") for i in confound_names) if not n else n
    tcompcor = any(i.startswith("t_comp_cor") for i in confound_names)
    if not cosine and (acompcor or tcompcor):
        LG.warning(
            "fMRIPrep applies high-pass filtering before running anatomical and temporal CompCor. It is "
            "recommended to include the discrete cosine-bases regressors ('cosine_XX') in `confound_names` "
            "if including any 'a_comp_cor' or 't_comp_cor' regressors."
        )



"""Internal function for checking the validity of parcel_approach."""

import copy, os, re

import numpy as np
from nilearn import datasets

from .pickle_utils import _convert_pickle_to_dict
from .logger import _logger

LG = _logger(__name__)

VALID_DICT_STUCTURES = {
    "Schaefer": {"n_rois": 400, "yeo_networks": 7, "resolution_mm": 1},
    "AAL": {"version": "SPM12"},
    "Custom": {
        "maps": "/location/to/parcellation.nii.gz",
        "nodes": ["LH_Vis1", "LH_Vis2", "LH_Hippocampus", "RH_Vis1", "RH_Vis2", "RH_Hippocampus"],
        "regions": {"Vis": {"lh": [0, 1], "rh": [3, 4]}, "Hippocampus": {"lh": [2], "rh": [5]}},
    },
}


def _check_parcel_approach(parcel_approach, call="TimeseriesExtractor"):
    """
    Pipeline to ensure ``parcel_approach`` is valid and process the ``parcel_approach`` if certain initialization
    keys are used.
    """
    if isinstance(parcel_approach, str) and parcel_approach.endswith(".pkl"):
        parcel_dict = _convert_pickle_to_dict(parcel_approach)
    else:
        parcel_dict = copy.deepcopy(parcel_approach)

    if not isinstance(parcel_dict, dict) or not list(parcel_dict)[0] in list(VALID_DICT_STUCTURES):
        error_message = (
            "Please include a valid `parcel_approach` in one of the following dictionary formats for "
            f"'Schaefer', 'AAL', or 'Custom': {VALID_DICT_STUCTURES}"
        )
        if not isinstance(parcel_dict, dict):
            raise TypeError(error_message)
        else:
            raise KeyError(error_message)

    if len(parcel_dict) > 1:
        raise ValueError(
            "Only one parcellation approach can be selected. Example format of `parcel_approach`:\n"
            f"{VALID_DICT_STUCTURES}"
        )

    if "Custom" in parcel_dict:
        # No return, simply validate structure
        _process_custom(parcel_dict, call)
    else:
        has_required_keys = _check_keys(parcel_dict)
        if not has_required_keys:
            parcel_dict = _process_aal(parcel_dict) if "AAL" in parcel_dict else _process_schaefer(parcel_dict)

    return parcel_dict


def _check_keys(parcel_dict):
    """
    Checks if the all required keys are in ``parcel_approach`` to skip processing of the ``parcel_approach``
    for "Schaefer" and "AAL".
    """
    required_keys = ["maps", "nodes", "regions"]

    return all(key in parcel_dict[list(parcel_dict)[0]] for key in required_keys)


def _process_aal(parcel_dict):
    """
    Converts a ``parcel_approach`` containing initialization keys for "AAL" the the final processed version.
    Uses nilearn to fetch the parcellation map and labels.
    """
    if "version" not in parcel_dict["AAL"]:
        LG.warning("'version' not specified in `parcel_approach`. Defaulting to 'SPM12'.")
        parcel_dict["AAL"].update({"version": "SPM12"})

    # Get atlas map; named "maps" to match nilearn and only contains a single file
    fetched_aal = datasets.fetch_atlas_aal(version=parcel_dict["AAL"]["version"], verbose=0)
    parcel_dict["AAL"].update({"maps": fetched_aal.maps})

    # Get nodes
    parcel_dict["AAL"].update({"nodes": list(fetched_aal.labels)})

    # Get regions
    regions = _collapse_aal_node_names(parcel_dict["AAL"]["nodes"])
    parcel_dict["AAL"].update({"regions": regions})

    # Clean initialization keys
    for key in VALID_DICT_STUCTURES["AAL"]:
        parcel_dict["AAL"].pop(key, None)

    return parcel_dict


def _process_custom(parcel_dict, call):
    """
    Ensures that "Custom" parcel approaches have the necessary keys. Performs basic validation before passing to
    ``_check_custom_structure``.
    """
    custom_example = {"Custom": VALID_DICT_STUCTURES["Custom"]}

    if call == "TimeseriesExtractor" and "maps" not in parcel_dict["Custom"]:
        raise ValueError(
            "For `Custom` parcel_approach, a nested key-value pair containing the key 'maps' with the "
            "value being a string specifying the location of the parcellation is needed. "
            f"Refer to example: {custom_example}"
        )

    if not all(check_subkeys := ["nodes" in parcel_dict["Custom"], "regions" in parcel_dict["Custom"]]):
        missing_subkeys = [["nodes", "regions"][x] for x, y in enumerate(check_subkeys) if y is False]
        error_message = f"The following subkeys haven't been detected {missing_subkeys}"

        if call == "TimeseriesExtractor":
            LG.warning(
                f"{error_message}. These labels are not needed for timeseries extraction but are needed "
                "for plotting."
            )
        else:
            raise ValueError(
                f"{error_message}. Certain subkeys are needed for plotting. Check the "
                "documentation for the required subkeys and reassign `parcel_approach` using "
                f"`self.parcel_approach`. Refer to the example structure:\n{custom_example}"
            )

    if call == "TimeseriesExtractor" and not os.path.isfile(parcel_dict["Custom"]["maps"]):
        raise FileNotFoundError(
            "The custom parcellation map does not exist in the specified file location: "
            f"{parcel_dict['Custom']['maps']}"
        )

    # Check structure
    _check_custom_structure(parcel_dict["Custom"], custom_example)


def _process_schaefer(parcel_dict):
    """
    Converts a ``parcel_approach`` containing initialization keys for "Schafer" the the final processed version.
    Uses nilearn to fetch the parcellation map and labels.
    """
    if "n_rois" not in parcel_dict["Schaefer"]:
        LG.warning("'n_rois' not specified in `parcel_approach`. Defaulting to 400 ROIs.")
        parcel_dict["Schaefer"].update({"n_rois": 400})

    if "yeo_networks" not in parcel_dict["Schaefer"]:
        LG.warning("'yeo_networks' not specified in `parcel_approach`. Defaulting to 7 networks.")
        parcel_dict["Schaefer"].update({"yeo_networks": 7})

    if "resolution_mm" not in parcel_dict["Schaefer"]:
        LG.warning("'resolution_mm' not specified in `parcel_approach`. Defaulting to 1mm.")
        parcel_dict["Schaefer"].update({"resolution_mm": 1})

    # Get atlas; named "maps" to match nilearn and only contains a single file
    fetched_schaefer = datasets.fetch_atlas_schaefer_2018(**parcel_dict["Schaefer"], verbose=0)
    parcel_dict["Schaefer"].update({"maps": fetched_schaefer.maps})
    network_name = "7Networks_" if parcel_dict["Schaefer"]["yeo_networks"] == 7 else "17Networks_"

    # Get nodes
    parcel_dict["Schaefer"].update(
        {"nodes": [label.decode().split(network_name)[-1] for label in fetched_schaefer.labels]}
    )

    # Get regions
    regions = dict.fromkeys([re.split("LH_|RH_", node)[-1].split("_")[0] for node in parcel_dict["Schaefer"]["nodes"]])
    parcel_dict["Schaefer"].update({"regions": list(regions)})

    # Clean initialization keys
    for key in VALID_DICT_STUCTURES["Schaefer"]:
        parcel_dict["Schaefer"].pop(key, None)

    return parcel_dict


def _check_custom_structure(custom_parcel, custom_example):
    """Validates the structure of the "nodes" and "regions" subkeys for user-defined ("Custom") parcel approaches."""
    example_msg = f"Refer to example: {custom_example}"

    if "nodes" in custom_parcel:
        if not (isinstance(custom_parcel.get("nodes"), (list, np.ndarray)) and list(custom_parcel.get("nodes"))):
            raise TypeError(
                "The 'nodes' subkey must be a non-empty list or numpy array containing the node labels. "
                f"{example_msg}"
            )

        if not all(isinstance(element, str) for element in custom_parcel["nodes"]):
            raise TypeError(
                "All elements in the 'nodes' subkey's list or numpy array must be a string. " f"{example_msg}"
            )

    if "regions" in custom_parcel:
        if not (custom_parcel.get("regions") and isinstance(custom_parcel.get("regions"), dict)):
            raise TypeError(
                "The 'regions' subkey must be a non-empty dictionary containing the regions/networks "
                f"labels. {example_msg}"
            )

        if not all(isinstance(key, str) for key in custom_parcel["regions"]):
            raise TypeError(f"All first level keys in the 'regions' subkey's dictionary must be strings. {example_msg}")

        regions = custom_parcel["regions"].values()
        if not all([all([hemisphere in region.keys() for hemisphere in ["lh", "rh"]]) for region in regions]):
            raise KeyError(
                "All second level keys in the 'regions' subkey's dictionary must contain 'lh' and 'rh'. "
                f"{example_msg}"
            )

        if not _check_custom_hemisphere_dicts(regions):
            raise TypeError(
                "Each 'lh' and 'rh' subkey in the 'regions' subkey's dictionary must contain a list of integers or "
                f"range of node indices. {example_msg}"
            )


def _check_custom_hemisphere_dicts(regions):
    """Ensures that the "Custom" parcel approach has the proper structure for the hemisphere dictionary."""
    # For the left and right hemisphere subkeys, check that they contain a list or range
    # Only check if each element of a list is an integer since range is guaranteed to be a sequence of integers already
    return all(
        isinstance(item[key], range)
        or (isinstance(item[key], list) and all(isinstance(element, int) for element in item[key]))
        for item in regions
        for key in ["lh", "rh"]
    )


def _collapse_aal_node_names(nodes, return_unique_names=True):
    """
    Creates general regions/networks from AAL labels (e.g. Frontal_Sup_2_L and Frontal_Inf_Oper_L -> Frontal) with
    special consideration for version "3v2".
    """
    # Names in "3v2" that don't split well .split("_")[0] or could be reduced more in the case of OFC, which
    # has OFCmed, OFCant, OFCpos
    special_names = ["N_Acc", "Red_N", "OFC"]

    collapsed_node_names = []

    for node in nodes:
        bool_vec = [name in node for name in special_names]

        if not any(bool_vec):
            collapsed_node_names.append(node.split("_")[0])
        else:
            indx = bool_vec.index(True)
            collapsed_node_names.append(special_names[indx])

    if return_unique_names:
        collapsed_node_names = list(dict.fromkeys(collapsed_node_names))

    return collapsed_node_names
